/**
 * Copyright (c) 2013-present, Facebook, Inc.
 * All rights reserved.
 *
 * This source code is licensed under the BSD-style license found in the
 * LICENSE file in the root directory of this source tree. An additional grant
 * of patent rights can be found in the PATENTS file in the same directory.
 */

'use strict'; // eslint-disable-line strict

module.exports = function (context) {
  var t = context.types;

  var idxRe = /\bidx\b/;

  function checkIdxArguments(file, node) {
    var args = node.arguments;
    if (args.length !== 2) {
      throw file.buildCodeFrameError(node, 'The `idx` function takes exactly two arguments.');
    }
    var arrowFunction = args[1];
    if (!t.isArrowFunctionExpression(arrowFunction)) {
      throw file.buildCodeFrameError(arrowFunction, 'The second argument supplied to `idx` must be an arrow function.');
    }
    if (!t.isExpression(arrowFunction.body)) {
      throw file.buildCodeFrameError(arrowFunction.body, 'The body of the arrow function supplied to `idx` must be a single ' + 'expression (without curly braces).');
    }
    if (arrowFunction.params.length !== 1) {
      throw file.buildCodeFrameError(arrowFunction.params[2] || arrowFunction, 'The arrow function supplied to `idx` must take exactly one parameter.');
    }
    var input = arrowFunction.params[0];
    if (!t.isIdentifier(input)) {
      throw file.buildCodeFrameError(arrowFunction.params[0], 'The parameter supplied to `idx` must be an identifier.');
    }
  }

  function makeCondition(node, state, inside) {
    if (inside) {
      return t.ConditionalExpression(t.BinaryExpression('!=', t.AssignmentExpression('=', state.temp, node), t.NullLiteral()), inside, state.temp);
    } else {
      return node;
    }
  }

  function makeChain(node, state, inside) {
    if (t.isCallExpression(node)) {
      return makeChain(node.callee, state, makeCondition(t.CallExpression(state.temp, node.arguments), state, inside));
    } else if (t.isMemberExpression(node)) {
      return makeChain(node.object, state, makeCondition(t.MemberExpression(state.temp, node.property, node.computed), state, inside));
    } else if (t.isIdentifier(node)) {
      if (node.name !== state.base.name) {
        throw state.file.buildCodeFrameError(node, 'The parameter of the arrow function supplied to `idx` must match ' + 'the base of the body expression.');
      }
      return makeCondition(state.input, state, inside);
    } else {
      throw state.file.buildCodeFrameError(node, 'The `idx` body can only be composed of properties and methods.');
    }
  }

  var idxVisitor = {
    CallExpression: function CallExpression(path, state) {
      var node = path.node;
      if (t.isIdentifier(node.callee) && node.callee.name === 'idx') {
        checkIdxArguments(state.file, node);
        var temp = path.scope.generateUidIdentifier('ref');
        var replacement = makeChain(node.arguments[1].body, {
          file: state.file,
          input: node.arguments[0],
          base: node.arguments[1].params[0],
          temp: temp
        });
        path.scope.push({ id: temp });
        path.replaceWith(replacement);
      }
    }
  };

  return {
    visitor: {
      Program: function Program(path, state) {
        // If there can't reasonably be an idx call, exit fast.
        if (path.scope.getOwnBinding('idx') || idxRe.test(state.file.code)) {
          // We're very strict about the shape of idx. Some transforms, like
          // "babel-plugin-transform-async-to-generator", will convert arrow
          // functions inside async functions into regular functions. So we do
          // our transformation before any one else interferes.
          path.traverse(idxVisitor, state);
        }
      }
    }
  };
};